class Const:
    CONNECT = 'connect'
    GET_SCHEMA = 'get_schema'
    UPDATE_SCHEMA = 'update_schema'
    CLASS_FRAME = 'class_frame'
    WOQL_SELECT = 'woql_select'
    GET_DOCUMENT = 'get_document'
    DELETE_DATABASE = 'delete_database'
    DELETE_DOCUMENT = 'delete_document'
    CREATE_DATABASE = 'create_database'
    CREATE_DOCUMENT = 'create_document'
    WOQL_UPDATE = 'woql_update'
